// Demo 2: fh.java

//  A File Demo -- A File Handler Utility
//  A Java Screen Based Application
//
//  by Charles W. Neville, 1998
//  Copyright Charles W. Neville, 1998

//  Java provides a complete suite of file handling classes and methods.  In 
//  addition to opening files, reading from and writing to files, and closing 
//  files, it is easy to "system things" like copy or rename files, delete 
//  files, get directory listings, and create or delete  directories.  In this 
//  demo, we shall see how to do all of these.
//
//  This demo will use command line arguments to see what to do.  Thus, to copy 
//  the file why.txt to the file whynot.txt, the user would type on the command 
//  line,
//
//      java fh -cp why.fil whynot.fil
//
//  This, by the way, is why we will violate the convention that class names 
//  always begin with a capital letter.  We want to use a simple two-letter 
//  commands, such as fh, as in UNIX.  We also follow the example of UNIX in 
//  using - before command line options, such as -cp for copy.
//
//  The complete set of user command examples is given below:
//
//      java fh -h                       display this message
//      java fh -cp why.fil whynot.fil   copy why.fil to whynot.fil
//      java fh -gp why.fil              get the complete path of why.fil
//      java fh -ls                      get a directory listing
//      java fh -md mydir                create (make) directory mydir
//      java fh -mv why.fil whynot.fil   rename (move) why.fil as whynot.fil
//      java fh -rm whynot.fil           remove (delete) whynot.fil
//
//  System things, such as getting directory listings, renaming files or 
//  deleting files, tend to be done with Java's class File.  An object of class 
//  File contains information (such as the file name) about a given physical 
//  file, together with methods for accessing or changing these.  However, an 
//  on object of class File does not contain any methods for doing file IO, 
//  that is for reading from or writing to files.  To do file IO, one uses 
//  various kinds of Java stream objects. 
//
//  Streams are the underlying abstraction behind almost all kinds of Java IO. 
//  There are many different kinds of streams in Java, and they may be used for 
//  many different purposes.  A stream is a simply a sequence of bytes flowing 
//  into or out of a Java object.  Although the idea of a stream is very 
//  simple, it is also very general.  A stream may be a sequence of bytes being 
//  read from or written to a physical file, or it may be a sequence of bytes 
//  flowing across a network.  For file IO, one typically use objects of the 
//  classes FileInputStream and FileOutputStream.

//  Review: remember, all Java applications, including screen based ones, must
//  have a public static void main method.

import java.io.*;  // have to import this every time we do file IO as well as
                   // keyboard IO in a screen based application.

public class fh {

    public final static int EOF = -1;  // end of file constant

    // array of proper usage messages - note use of initializer list
    private final String usageMessage[] =  
            { "Usage:",
              "java fh -cp file1 file2  copy file1 to file2",
              "java fh -gp file1        get path of file1",
              "java fh -h               display this message",
              "java fh -ls              get a directory listing",
              "java fh -md dir1         make directory dir1",
              "java fh -mv file1 file2  rename (move) file1 as file2",
              "java fh -rm file1        remove file1" };

    public static void main(String args[]) {

        File newDirectory;  // create a new subdirectory of current directory;
        String newPath;

        fh f = new fh();  // have to create an object because we
            // reference (non-static) object methods and variables in main
 
        if (args.length == 0) {
           f.printUsageMessage();
           return;  // bail out of program
        }
        // else test which option -- can't use switch because strings are not
        //    a primitive data type.
        // remember, NEVER use == to test for equality of string content,
        // ALWAYS use equals().
        else if (args[0].equals("-cp")) {
           // be sure there are two more file arguments and copy files           
           if (args.length != 3) {
               f.printUsageMessage();
               return;
           }
           // note: don't need else here because of return above

           // cp returns true if the copy operation succeeded, false
           // otherwise.  So bury the call in an if so we can print
           // an error message if the operation did not succeed.           
           if (!f.cp(args[1], args[2])) {
               System.out.println("File could not be copied.");
           }
        }
        else if (args[0].equals("-gp")) {
           // be sure there is one file argument and get path
           if (args.length != 2) {
              f.printUsageMessage();
              return;
           }
           
           System.out.println(f.gp(args[1]));
        }
        else if (args[0].equals("-h")) {
           // don't need to test argument list because we will print the
           // usageMessage anyway.
           f.printUsageMessage();
        }
        else if (args[0].equals("-ls")) {
           // be sure the option argument is the only one and print directory 
           // listing
           if (args.length != 1) {
              f.printUsageMessage();
              return;
           }
           f.printStringArray(f.ls());
        }
        else if (args[0].equals("-md")) {
          // be sure there is only one directory argument and make the new 
          // directory.
          if (args.length != 2) {
             f.printUsageMessage();
             return;
          }
          // md returns true if the make directory operation succeeded,
          // false otherwise.  So bury the call in an if so we can print
          // an error message if the operation did not succeed.
          if (!f.md(args[1])) {
             System.out.println("Directory could not be created.");
          }
        }
        else if (args[0].equals("-mv")) {
          // be sure there are two file arguments and move (rename) the file
          if (args.length != 3) {
             f.printUsageMessage();
             return;
          }
          if (!f.mv(args[1], args[2])) {
            System.out.println("File could not be moved (renamed)");
          }
        }
        else if (args[0].equals("-rm")) {
          // be sure there is only one file argument and remove the file
          if (args.length != 2) {
             f.printUsageMessage();
             return;
          }
          if (!f.rm(args[1])) {
             System.out.print("File (directory) could not be removed ");
             System.out.println("(deleted)");
          }
        }
        else {
          // there was an incorrect option
          f.printUsageMessage();
          return;
        }
 
        return;        

    } // main

    // various utility methods

    // print an array of strings, such as the usageMessage
    // or a directory listing
    private void printStringArray(String s[]) {

       for(int i = 0; i < s.length; i++) {
          System.out.println(s[i]);
       }
    } // printStringArray

    // print usage message
    private void printUsageMessage() {

       printStringArray(usageMessage);
    } // printUsageMessage

    // print security error message
    private void printSecurityError() {

       System.out.println("Insufficient privileges for attempted operation");
    } // printSecurityError

    // actual file handling methods

    // copy file1 to file2
    public boolean cp(String file1, String file2) {

        FileInputStream inStream;     // read from inStream
        FileOutputStream outStream;   // write to outStream

        int b;  // currently read byte stored as an int in range 0 -- 255.

        // constructing a FileInputStream or OutputStream object throws a
        // SecurityException and a IOException (because the files must be
        // opened and/or created), so we have to use try -- catch clauses
        // to catch the exceptions.
        try {
           inStream = new FileInputStream(file1);   // this creates a
               // FileInputStream attached to the physical file named file1, 
               // and it opens file1 for reading.
        }
        catch (SecurityException e) {
            printSecurityError();
            return false;  // copy operation failed
        }
        catch (IOException e) {
            System.out.println("Error opening input file.");
            return false;  // copy operation failed
        }

        try {
            outStream = new FileOutputStream(file2); // this creates a
               // FileOutputStream attached to the physical file named file2,
               // and it opens file2 for writing. If file2 does not exist, 
               // it creates a new file.
        }
        catch (SecurityException e) {
            printSecurityError();
            return false;   // copy operation failed
        }
        catch (IOException e) {
            System.out.println("Error opening output file.");
            return false;   // copy operation failed
        }
       
        // The file copy algorithm is simple:
        //
        //   while (there is more data) {
        //      read a byte of data from the input file;
        //      write the byte of data to the output file;
        //   }
        //
        //  To do this, we could simply use the FileInputStream and 
        //  FileOutputStream directly, as follows:
        // 
        //    while ((b = inStream.read()) != EOF) {
        //      outStream.write(b);
        //    }
        //
        //  But buffered IO is MUCH more efficient, so we shall use a
        //  BufferedInputStream and a BufferedOutputStream to do the actual
        //  copying.  You use a BufferedInputStream by attaching it to a
        //  FileInputStream (or, in more general cases, to some other kind
        //  of InputStream), and similarly with a BufferedOutputStream.

        // construct a BufferedInputStream and BufferedOutputStream attached,
        // respectively, to the FileInputStream inStream and the 
        // FileOutputStream outStream.
        BufferedInputStream inBufStream = new BufferedInputStream(inStream);
        BufferedOutputStream outBufStream = 
                                          new BufferedOutputStream(outStream);

        // do the actual copying.  Since the read and write methods throw 
        // IOExceptions, we have to use a try -- catch clause again.
        try {
           // copy byte by byte -- buffered IO for efficiency
           while ((b = inBufStream.read()) != EOF) {
               outBufStream.write(b);  // this is inherited from C and is a
                  // little confusing.  We want to read and write a byte (8 
                  // bits), but we seem to be reading and writing a 32 bit 
                  // int.  Java reads a byte as a 32 bit int in the range 0 
                  // -- 255, and then writes it back as a byte.  Java, like C, 
                  // uses ints so the read() method can return -1, which is 
                  // NOT an int in the range 0 -- 255, to mark the end of file.
            }
            // close the streams -- closing automatically flushes the buffers
            inBufStream.close();   inStream.close();
            outBufStream.close();  outStream.close();
            return true;    // copy operation succeeded
        }
        catch (IOException e) {
            System.out.println("Error reading from or writing to file");
             return false;   // copy operation failed
        } 
    }  // cp

    // get the path of file with name fileName
    public String gp(String fileName) {

       // The Java File class includes methods for getting information
       // about files, such as a file's path, name, parent directory.  It also 
       // contains methods for deleting and renaming files, and even making
       // new directories.  Whenever you want to do one of these things, you
       // always construct a File object first and then call the appropriate
       // File class method.

       // Note: The Java File class does NOT contain methods for reading 
       // information from files, or writing information to files.  To do this, 
       // you use objects of class FileInputStream or FileOutputStream.

       // construct a Java File object and get its path.
       File file = new File(fileName);

       // almost every File object method throws a SecurityException, so
       // we have to use try -- catch clauses to catch any security exception.
       try {
          return file.getPath();  // return the path
       }
       catch (SecurityException e) {
          printSecurityError();
          return "";
       }
    } // gp

    // get a directory listing of the current directory
    public String[] ls() {

       String nullReturn[] = {""};  // something to return if can't get
                                    // a directory listing

       File cwd = new File(".");  // current working directory

       try {
          return cwd.list();     // return the directory listing
       }
       catch (SecurityException e) {
          printSecurityError();
          return nullReturn;
       }
    }  // ls

    // make a subdirectory with name dirName in the current directory
    public boolean md(String dirName) {

       File dir = new File(dirName);  // directory object now exists in Java
                  // with the given name (a directory is ALWAYS a special kind 
                  // the directory does not exist yet in the operating system.
       
       try {
          return dir.mkdir(); // make the directory in the operating system
             // mkdir() returns true if the operation was successful,
             // false otherwise.
       }
       catch (SecurityException e) {
          printSecurityError();
          return false;
       }
    } // md

    // move (rename) file with fileName1 as fileName2
    public boolean mv(String fileName1, String fileName2) {

       File file1 = new File(fileName1);
       File file2 = new File(fileName2);  // File method renameTo takes a
           // File object as an argument, so we have to create the file2 
           // object.  Remember, we haven't created any files or directories
           // yet from the operating sytems point of view, we have just
           // created a second Java File object to hold file information.

       try {
           return file1.renameTo(file2);
       }
       catch (SecurityException e) {
           printSecurityError();
           return false;
       }      
    } // mv

    // remove (delete) file with name fileName
    public boolean rm(String fileName) {

       File file = new File(fileName);
       try {
          return file.delete();
       }
       catch (SecurityException e) {
          printSecurityError();
          return false;
       }
    }                                          

}  // class fh
